import { Component, ComponentConstructor } from 'preact';
import { Dispatcher } from './dispatcher';
import { Cmd, isCmdDispatch } from './cmd';

export interface Message<K, V = any, E extends Error = any> {
    kind: K;
    value: V;
    error: E;
}

export interface ComponentDefinition<Model, Msg extends Message<any>, Props> {
    update(model: Model | undefined, msg: Msg): Model | [Model, Cmd<Msg['kind']>];
    view(model: Model, dispatch: Dispatcher<Msg['kind']>): JSX.Element | null;

    init?(model: Model, dispatch: Dispatcher<Msg['kind']>): void;
    props?(props: Props, dispatch: Dispatcher<Msg['kind']>): void;
}

function runUpdate<Model, Msg extends Message<any>>(
    comp: ComponentDefinition<Model, Msg, any>,
    dispatch: Dispatcher<Msg['kind']>,
    model: Model,
    msg: Msg,
): Model {
    const res = comp.update(model, <any>msg);
    let newModel: Model;
    if (isCmdDispatch<Model, Msg['kind']>(res)) {
        newModel = res[0];
        res[1](dispatch);
    } else {
        newModel = res;
    }

    return newModel;
}

export function component<Model, Msg extends Message<any>, Props = {}>(
    comp: ComponentDefinition<Model, Msg, Props>,
): ComponentConstructor<Props, Model> {
    return class extends Component<Props, Model> {
        dispatchers = new Map<Msg['kind'], (value?: any, err?: Error) => void>();

        constructor(...args: any[]) {
            super(...args);

            const state = runUpdate(comp, this.dispatch, undefined, {
                kind: -1,
                value: undefined,
                error: undefined,
            } as any)!;

            // Clone the initial state onto the state object
            // to avoid preact from mutating the initial state
            // on later calls to setState
            this.state = { ...(state as any) };
        }

        componentWillMount() {
            if (comp.props != null) {
                comp.props(this.props, this.dispatch);
            }
        }

        componentDidMount() {
            if (comp.init != null) {
                comp.init(this.state, this.dispatch);
            }
        }

        componentWillReceiveProps(props: Props) {
            // TODO: only if changed?
            if (comp.props != null) {
                comp.props(props, this.dispatch);
            }
        }

        dispatch = (kind: Msg['kind']) => {
            let dispatcher = this.dispatchers.get(kind);
            if (dispatcher == null) {
                dispatcher = (value: any, err: Error) => {
                    let msg = { kind, value, error: err };
                    if (value instanceof Event && value.target instanceof HTMLInputElement) {
                        switch (value.target.type) {
                            case 'checkbox':
                                msg = { ...msg, value: value.target.checked };
                                break;
                            default:
                                msg = { ...msg, value: value.target.value };
                                break;
                        }
                    }

                    this.replaceState(runUpdate(comp, this.dispatch, this.state, msg as any));
                };
                this.dispatchers.set(kind, dispatcher);
            }

            return dispatcher;
        };

        // Copied from preact-compat
        replaceState(state: Model, callback?: () => void) {
            this.setState(state, callback);
            for (let i in this.state) {
                if (!(i in state)) {
                    delete this.state[i];
                }
            }
        }

        render() {
            return comp.view(this.state, this.dispatch);
        }
    };
}
